/*
DevicePathUtil by joevt
*/

#undef __pic__

	#include "UefiDevicePathLib.h"

	#include <ctype.h>
	#include <stdio.h>
	#include <stdlib.h>
	#include <string.h>
	#include <stdarg.h>
	#include <unistd.h>
	#include <wchar.h>

//========================================================================================

EFI_SYSTEM_TABLE ST;
EFI_BOOT_SERVICES BS;
EFI_SYSTEM_TABLE *gST = NULL;
EFI_BOOT_SERVICES *gBS = NULL;

#define EFI_UART_DEVICE_PATH_GUID \
  { \
    0x37499a9d, 0x542f, 0x4c89, {0xa0, 0x26, 0x35, 0xda, 0x14, 0x20, 0x94, 0xe4 } \
  }


EFI_GUID gEfiDebugPortProtocolGuid = EFI_DEBUGPORT_PROTOCOL_GUID;
EFI_GUID gEfiPcAnsiGuid = EFI_PC_ANSI_GUID;
EFI_GUID gEfiPersistentVirtualCdGuid = EFI_PERSISTENT_VIRTUAL_CD_GUID;
EFI_GUID gEfiPersistentVirtualDiskGuid = EFI_PERSISTENT_VIRTUAL_DISK_GUID;
EFI_GUID gEfiSasDevicePathGuid = EFI_SAS_DEVICE_PATH_GUID;
EFI_GUID gEfiUartDevicePathGuid = EFI_UART_DEVICE_PATH_GUID;
EFI_GUID gEfiVirtualCdGuid = EFI_VIRTUAL_CD_GUID;
EFI_GUID gEfiVirtualDiskGuid = EFI_VIRTUAL_DISK_GUID;
EFI_GUID gEfiVT100Guid = EFI_VT_100_GUID;
EFI_GUID gEfiVT100PlusGuid = EFI_VT_100_PLUS_GUID;
EFI_GUID gEfiVTUTF8Guid = EFI_VT_UTF8_GUID;

EFI_GUID gEfiDevicePathProtocolGuid = EFI_DEVICE_PATH_PROTOCOL_GUID;


VOID *
InternalAllocateCopyPool (
   UINTN            AllocationSize,
   CONST VOID       *Buffer
  )
{
  VOID  *Memory;

  ASSERT (Buffer != NULL);
//  ASSERT (AllocationSize <= (MAX_ADDRESS - (UINTN) Buffer + 1));

  Memory = malloc (AllocationSize);
  if (Memory != NULL) {
     Memory = memcpy (Memory, Buffer, AllocationSize);
  }
  return Memory;
}


VOID *
AllocateCopyPool (
   UINTN       AllocationSize,
   CONST VOID  *Buffer
  )
{
  return InternalAllocateCopyPool (AllocationSize, Buffer);
}


EFI_STATUS
EFIAPI
DontHandleProtocol (
  IN EFI_HANDLE       UserHandle,
  IN EFI_GUID         *Protocol,
  OUT VOID            **Interface
  )
{
	return EFI_UNSUPPORTED;
}


EFI_STATUS
EFIAPI
UefiBootServicesTableLibConstructor (
  )
{
	ZeroMem(&BS, sizeof(&BS));
	BS.HandleProtocol = &DontHandleProtocol;
	
	ZeroMem(&ST, sizeof(&ST));
	ST.BootServices = &BS;

	//
	// Cache pointer to the EFI System Table
	//
	gST = &ST;

	//
	// Cache pointer to the EFI Boot Services Table
	//
	gBS = gST->BootServices;

	return EFI_SUCCESS;
}


STATIC
VOID
PrintMem (
          CONST VOID *Buffer,
          UINTN      Count
          )
{
    CONST UINT8 *Bytes;
    UINTN       Idx;
    
    Bytes = Buffer;
    for (Idx = 0; Idx < Count; Idx++) {
        printf("%02x", Bytes[Idx]);
    }
}


//========================================================================================
const char program_name[] = "DevicePathUtil";

static char options[] = "dvs" ;

static char help_msg[] =
"# Usage: DevicePathUtil [-v] [-d] [-s] [efidevicepath...]\n"
"# -v    Verbose\n"
"# -d    DisplayOnly - shorter text representation of the display node is used, where applicable\n"
"# -s    AllowShortcuts - shortcut forms of text representation for a device node can be used, where applicable\n"
"";

int gOptVerbose = FALSE;
int gOptAllowShortCuts = FALSE;
int gOptDisplayOnly = FALSE;

//========================================================================================


int islowerhex(char c)
{
	return ((c >= '0' && c <= '9') || (c >= 'a' && c <= 'f'));
}


int ishex(char c)
{
	return ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'F') || (c >= 'a' && c <= 'f'));
}


int hexchar(char c)
{
	return (c - (c < 'A' ? '0' : ((c < 'a' ? 'A' : 'a') - 10)));
}


int hextobyte(char* c)
{
	return (hexchar(*c) << 4) + hexchar(c[1]);
}


int allhex(char *arg, unsigned long len)
{
	for (int i = 0; i < len; i++)
		if (!ishex(arg[i]))
			return 0;
	return 1;
}


int DevicePathUtilToText(void* bytepath, int bytepathlen)
{
    int result = 0;
	if (gOptVerbose)
	{
		fprintf(stdout, "# Converting %d bytes to text\n", bytepathlen);
	}

	CHAR16* textpath = ConvertDevicePathToText(bytepath, gOptDisplayOnly, gOptAllowShortCuts);
    
    if (textpath)
    {
        UINTN textlen = StrLen(textpath);

        if (gOptVerbose)
        {
            fprintf(stdout, "# Text path %llu\n", textlen);
        }

        CHAR8* asciitextpath = AllocatePool(textlen);
        if (asciitextpath)
        {
            UnicodeStrToAsciiStr(textpath, asciitextpath);
            fprintf(stdout, "%s\n", asciitextpath);
            FreePool(asciitextpath);
        }
        else
        {
            result = 1;
        }
        FreePool(textpath);
    }
    else
    {
        result = 1;
    }
    return result;
}


int DevicePathUtilFromText(void* asciitextpath, int asciitextpathlen)
{
    int result = 0;
    if (gOptVerbose)
    {
        fprintf(stdout, "# Converting %d characters to bytes\n", asciitextpathlen);
    }
    
    CHAR16* textpath = AllocatePool(asciitextpathlen * sizeof (CHAR16));
    if (textpath)
    {
        AsciiStrToUnicodeStr(asciitextpath, textpath);
        EFI_DEVICE_PATH_PROTOCOL *bytepath = ConvertTextToDevicePath(textpath);
        if (bytepath)
        {
            int bytepathlen = 0;
            for (EFI_DEVICE_PATH_PROTOCOL *node = bytepath;;)
            {
                int nodelen = node->Length[0] | node->Length[1] << 8;
                bytepathlen += nodelen;
                if (node->Type == END_DEVICE_PATH_TYPE && node->SubType == END_ENTIRE_DEVICE_PATH_SUBTYPE) break;
                node = (EFI_DEVICE_PATH_PROTOCOL *)((UINT8*)node + nodelen);
            }
            PrintMem (bytepath, bytepathlen);
            putchar('\n');
        }
        else
        {
            result = 1;
        }
        FreePool(textpath);
    }
    else
    {
        result = 1;
    }

    return result;
}


int parse_generic_option(char *arg, unsigned long arglen)
{
	if (gOptVerbose)
	{
		fprintf(stdout, "# Processing argument of length %lu\n", arglen);
	}

	int i;

	int bytepathlenmax = 10000;
	int bytepathlen = 0;
	UINT8* bytepath = malloc(bytepathlenmax);
	if (!bytepath) return 1;
	
	if (!arg || !arglen)
	{
		return 1;
	}

	if (arg[0] == '%')
	{
        // convert hex "nvram -p" path
        for (i = 0; i < arglen;)
		{
			if (bytepathlen == bytepathlenmax)
			{
				bytepathlenmax += 10000;
				bytepath = realloc(bytepath, bytepathlenmax);
				if (!bytepath)
					return 1;
			}
			if (arg[i] == '%' && i+2 < arglen && islowerhex(arg[i+1]) && islowerhex(arg[i+2]))
			{
                i++;
				bytepath[bytepathlen++] = hextobyte(arg + i);
				i += 2;
			}
			else
			{
				bytepath[bytepathlen++] = arg[i++];
			}
		}
		return DevicePathUtilToText(bytepath, bytepathlen);
	}
	else if ((arglen & 1) == 0 && allhex(arg, arglen))
	{
        // convert hex "xxd -p" path
		for (i = 0; i < arglen; i += 2)
		{
            if (bytepathlen == bytepathlenmax)
            {
                bytepathlenmax += 10000;
                bytepath = realloc(bytepath, bytepathlenmax);
                if (!bytepath)
                    return 1;
            }
			bytepath[bytepathlen++] = hextobyte(arg + i);
		}
		return DevicePathUtilToText(bytepath, bytepathlen);
	}
	else if (!isprint(arg[0]))
	{
        // convert binary path
		return DevicePathUtilToText(arg, arglen);
	}
	else
	{
		return DevicePathUtilFromText(arg, arglen);
	}
	
	return 0;
}

int main(int argc, char **argv)
{
	UefiBootServicesTableLibConstructor();
	
	int i;
	int gotParam = 0; 
	
	while ((i = getopt(argc, argv, options)) != -1)
	{
	    switch (i)
	    {
	    	case 'v': gOptVerbose = TRUE; break;
	    	case 's': gOptAllowShortCuts = TRUE; break;
	    	case 'd': gOptDisplayOnly = TRUE; break;
	    	
			case '?':
				if (isprint (optopt))
					fprintf (stderr, "# Unknown option `-%c'.\n", optopt);
				else
					fprintf (stderr, "# Unknown option character `\\x%x'.\n", optopt);
				goto bad;
				
			default:
				fprintf(stderr, "# Unknown option\n");
bad:
				fprintf(stderr, "%s", help_msg);
				return 1;
		}
	}
	
/*
    if (gOptVerbose)
	{
		fprintf(stdout, "# Testing character sizes\n");
		wchar_t testw[100] = L"12";
		char* test = (char*)testw;
		fprintf(stdout, "# %lu %lu %02x%02x%02x%02x\n", sizeof(test[0]), sizeof(testw[0]), test[0], test[1], test[2], test[3]); // expected: 1 2 31003200
	}
*/

	for (i = optind; i < argc; i++)
	{
		parse_generic_option(argv[i], strlen(argv[i]));
        i++;
		gotParam = 1;
	}
	
	if (!gotParam)
	{
		if (gOptVerbose) {
			fprintf(stdout, "# Reading from stdin\n");
		}
	
		unsigned long argtlenmax = 10000;
		char* argt = malloc(argtlenmax);
		if (!argt)
			return 1;
		argt[0] = '\0';
		unsigned long argtlen = 0;
		for (;;)
		{
			if (argtlen == argtlenmax)
			{
				argtlenmax += 10000;
				argt = realloc(argt, argtlenmax);
				if (!argt)
					return 1;
			}
			unsigned long readlen = fread(argt + argtlen, 1, argtlenmax - argtlen, stdin);
			if (!readlen)
				break;
			argtlen += readlen;
		}
		parse_generic_option(argt, argtlen);
	}
		
	return 0;
}
